# -*- coding: utf-8 -*-

import warnings
from collections.abc import Callable

from django.conf import settings
from django.utils.module_loading import import_string


def su_login_callback(user):
    if hasattr(settings, "SU_LOGIN"):
        warnings.warn(
            "SU_LOGIN is deprecated, use SU_LOGIN_CALLBACK",
            DeprecationWarning,
        )

    func = getattr(settings, "SU_LOGIN_CALLBACK", None)
    if func is not None:
        if not isinstance(func, Callable):
            func = import_string(func)
        return func(user)
    return user.has_perm("auth.change_user")


def custom_login_action(request, user):
    func = getattr(settings, "SU_CUSTOM_LOGIN_ACTION", None)
    if func is None:
        return False

    if not isinstance(func, Callable):
        func = import_string(func)
    func(request, user)

    return True



# -*- coding: utf-8 -*-

from ajax_select import LookupChannel, register
from django.db.models import Q
from django.utils.encoding import force_text
from django.utils.html import escape

from django_su import get_user_model


@register("django_su")
class UsersLookup(LookupChannel):

    model = get_user_model()

    def get_query(self, q, request):
        return self.model.objects.filter(
            Q(username__icontains=q) | Q(pk__icontains=q)
        ).order_by("pk")

    def format_match(self, obj):
        return escape(
            force_text("%s [pk: %s]" % (obj.get_full_name() or obj.username, obj.pk))
        )

    def format_item_display(self, obj):
        return escape(
            force_text("%s [pk: %s]" % (obj.get_full_name() or obj.username, obj.pk))
        )



# -*- coding: utf-8 -*-

from django.contrib.auth import BACKEND_SESSION_KEY, SESSION_KEY, get_user_model


try:
    from django.contrib.auth import HASH_SESSION_KEY
except ImportError:
    HASH_SESSION_KEY = "_auth_user_hash"

User = get_user_model()


def su_login_callback(user):
    if user.is_active and user.is_staff:
        return True
    return user.has_perm("auth.change_user")


def _get_user_session_key(request):
    # This value in the session is always serialized to a string, so we need
    # to convert it back to Python whenever we access it.

    return User._meta.pk.to_python(request.session[SESSION_KEY])


def custom_login(request, user):
    session_auth_hash = ""
    if user is None:
        user = request.user
    if hasattr(user, "get_session_auth_hash"):
        session_auth_hash = user.get_session_auth_hash()

    if SESSION_KEY in request.session:
        if _get_user_session_key(request) != user.pk or (
            session_auth_hash
            and request.session.get(HASH_SESSION_KEY) != session_auth_hash
        ):
            # To avoid reusing another user's session, create a new, empty
            # session if the existing session corresponds to a different
            # authenticated user.
            request.session.flush()
    else:
        request.session.cycle_key()
    request.session[SESSION_KEY] = user._meta.pk.value_to_string(user)
    request.session[BACKEND_SESSION_KEY] = user.backend
    request.session[HASH_SESSION_KEY] = session_auth_hash
    if hasattr(request, "user"):
        request.user = user

    try:
        from django.middleware.csrf import rotate_token

        rotate_token(request)
    except ImportError:
        pass



"""
Django settings for app project.

For more information on this file, see
https://docs.djangoproject.com/en/1.7/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.7/ref/settings/
"""

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
import os


BASE_DIR = os.path.dirname(os.path.dirname(__file__))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.7/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "YOUR_SECRET_KEY"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

TEMPLATE_DEBUG = True

ALLOWED_HOSTS = []

TEMPLATE_DIRS = (os.path.join(os.path.dirname(__file__), "templates"),)

TEMPLATE_CONTEXT_PROCESSORS = (
    "django.core.context_processors.tz",
    "django.core.context_processors.i18n",
    "django.core.context_processors.media",
    "django.core.context_processors.static",
    "django.core.context_processors.request",
    "django.contrib.auth.context_processors.auth",
    "django.core.context_processors.debug",
    "django_su.context_processors.is_su",
)

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [
            os.path.join(os.path.dirname(__file__), "templates"),
        ],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.i18n",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.template.context_processors.debug",
                "django.contrib.messages.context_processors.messages",
                "django_su.context_processors.is_su",
            ],
        },
    },
]

# Application definition

MIDDLEWARE = [
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
]

PROJECT_APPS = [
    "django_su",
]

INSTALLED_APPS = [
    # 'suit',  # pip install django-suit
    "django.contrib.auth",
    "django.contrib.sites",
    "django.contrib.sessions",
    "django.contrib.staticfiles",
    "django.contrib.contenttypes",
    "django.contrib.admin",
    "django.contrib.messages",
    # 'guardian',
    "formadmin",  # pip install django-form-admin
    "ajax_select",  # pip install django-ajax-selects
]

INSTALLED_APPS = PROJECT_APPS + INSTALLED_APPS

ROOT_URLCONF = "example.urls"

SITE_ID = 1

# Database
# https://docs.djangoproject.com/en/1.7/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": os.path.join(BASE_DIR, "db.sqlite3"),
    }
}

# Internationalization
# https://docs.djangoproject.com/en/1.7/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_L10N = True

USE_TZ = True

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.7/howto/static-files/

STATIC_URL = "/static/"

AUTHENTICATION_BACKENDS = (
    "django.contrib.auth.backends.ModelBackend",
    # "guardian.backends.ObjectPermissionBackend",
    "django_su.backends.SuBackend",
)

# ANONYMOUS_USER_ID = -1

# URL to redirect after the login.
# Default: "/"
SU_LOGIN_REDIRECT_URL = "/"

# URL to redirect after the logout.
# Default: "/"
SU_LOGOUT_REDIRECT_URL = "/"

# A function to specify the perms that the user must have can use django_su
# Default: None
SU_LOGIN_CALLBACK = "example.utils.su_login_callback"

# A function to override the django.contrib.auth.login(request, user)
# function so you can set session data, etc.
# Default: None
SU_CUSTOM_LOGIN_ACTION = "example.utils.custom_login"

if "ajax_select" in INSTALLED_APPS:
    AJAX_LOOKUP_CHANNELS = {
        "django_su": ("example.lookups", "UsersLookup"),
    }



# -*- coding: utf-8 -*-

from django import forms
from django.conf import settings
from django.contrib.auth import get_user_model
from django.utils.translation import gettext_lazy as _


User = get_user_model()


class UserSuForm(forms.Form):

    username_field = getattr(User, "USERNAME_FIELD", None)

    user = forms.ModelChoiceField(
        label=_("Users"),
        queryset=User._default_manager.order_by(username_field)
        if username_field
        else User._default_manager.all(),
        required=True,
    )  # pylint: disable=W0212

    use_ajax_select = False

    def __init__(self, *args, **kwargs):
        super(UserSuForm, self).__init__(*args, **kwargs)

        if "ajax_select" in settings.INSTALLED_APPS and getattr(
            settings, "AJAX_LOOKUP_CHANNELS", None
        ):
            from ajax_select.fields import AutoCompleteSelectField

            lookup = settings.AJAX_LOOKUP_CHANNELS.get("django_su", None)
            if lookup is not None:
                old_field = self.fields["user"]

                self.fields["user"] = AutoCompleteSelectField(
                    "django_su",
                    required=old_field.required,
                    label=old_field.label,
                )
                self.use_ajax_select = True

    def get_user(self):
        return self.cleaned_data.get("user", None)

    def __str__(self):
        if "formadmin" in settings.INSTALLED_APPS:
            try:
                from formadmin.forms import as_django_admin

                return as_django_admin(self)
            except ImportError:
                pass
        return super(UserSuForm, self).__str__()



# -*- coding: utf-8 -*-

from django.urls import path

from .views import login_as_user, su_login, su_logout


urlpatterns = [
    path("", su_logout, name="su_logout"),
    path("login/", su_login, name="su_login"),
    path("<int:user_id>/", login_as_user, name="login_as_user"),
]



from django.contrib.auth import get_user_model
from django.test import TestCase


User = get_user_model()


class TestSuBackend(TestCase):
    def setUp(self):
        super(TestSuBackend, self).setUp()
        from django_su.backends import SuBackend

        self.user = User.objects.create(username="testuser")
        self.backend = SuBackend()

    def test_authenticate_do_it(self):
        """Ensure authentication passes when su=True and user id is valid"""
        self.assertEqual(
            self.backend.authenticate(su=True, user_id=self.user.pk), self.user
        )

    def test_authenticate_dont_do_it(self):
        """Ensure authentication fails when su=False and user id is valid"""
        self.assertEqual(
            self.backend.authenticate(su=False, user_id=self.user.pk), None
        )

    def test_authenticate_id_none(self):
        """Ensure authentication fails when user_id is None"""
        self.assertEqual(self.backend.authenticate(su=True, user_id=None), None)

    def test_authenticate_id_non_existent(self):
        """Ensure authentication fails when user_id doesn't exist"""
        self.assertEqual(self.backend.authenticate(su=True, user_id=999), None)

    def test_authenticate_id_invalid(self):
        """Ensure authentication fails when user_id is invalid"""
        self.assertEqual(self.backend.authenticate(su=True, user_id="abc"), None)

    def test_get_user_exists(self):
        """Ensure get_user returns the expected user"""
        self.assertEqual(self.backend.get_user(user_id=self.user.pk), self.user)

    def test_get_user_does_not_exist(self):
        """Ensure get_user returns None if user is not found"""
        self.assertEqual(self.backend.get_user(user_id=999), None)



from datetime import date, timezone

from django.conf import settings
from django.contrib import auth
from django.contrib.auth import get_user_model
from django.contrib.sessions.backends import cached_db
from django.test import Client, TestCase
from django.urls import reverse
from django.utils.datetime_safe import datetime


User = get_user_model()


class SuViewsBaseTestCase(TestCase):
    def setUp(self):
        super(SuViewsBaseTestCase, self).setUp()
        from django_su.views import login_as_user

        self.authorized_user = self.user("authorized", is_superuser=True)
        self.unauthorized_user = self.user("unauthorized")
        self.destination_user = self.user("destination")
        self.view = login_as_user
        self.client = self.make_client()
        # Causes errors with validation.
        # TODO: Investigate
        if "ajax_select" in settings.INSTALLED_APPS:
            settings.INSTALLED_APPS.remove("ajax_select")

    def user(self, username, **kwargs):
        user = User.objects.create(username=username, **kwargs)
        user.set_password("pass")
        user.save()
        return user

    def make_client(self):
        client = Client()
        s = cached_db.SessionStore()
        s.save()
        client.cookies[settings.SESSION_COOKIE_NAME] = s.session_key
        return client


class LoginAsUserViewTestCase(SuViewsBaseTestCase):
    def test_login_success(self):
        """Ensure login works for a valid user"""
        self.client.login(username="authorized", password="pass")
        response = self.client.post(
            reverse("login_as_user", args=[self.destination_user.id])
        )
        self.assertEqual(response.status_code, 302)
        # Check the user is logged in in the session
        self.assertIn(auth.SESSION_KEY, self.client.session)
        self.assertEqual(
            str(self.client.session[auth.SESSION_KEY]), str(self.destination_user.id)
        )
        # Check the 'exit_users_pk' is set so we know which user to change back to
        self.assertIn("exit_users_pk", self.client.session)
        pk, backend = self.client.session["exit_users_pk"][0]
        self.assertEqual(str(pk), str(self.authorized_user.pk))
        self.assertEqual(backend, "django.contrib.auth.backends.ModelBackend")

    def test_login_user_id_invalid(self):
        """Ensure login fails with an invalid user id"""
        self.client.login(username="authorized", password="pass")
        response = self.client.post("/su/abc/")
        self.assertEqual(response.status_code, 404)
        # User should still be logged in, but as the original user
        self.assertIn(auth.SESSION_KEY, self.client.session)
        self.assertEqual(
            str(self.client.session[auth.SESSION_KEY]), str(self.authorized_user.id)
        )
        # Exit user should never get set
        self.assertNotIn("exit_users_pk", self.client.session)

    def test_login_without_permission(self):
        """Ensure login fails when the current user lacks permission"""
        self.client.login(username="unauthorized", password="pass")
        with self.settings(SU_LOGIN_CALLBACK=None):
            response = self.client.post(
                reverse("login_as_user", args=[self.destination_user.id])
            )
        self.assertEqual(response.status_code, 302)
        # User should still be logged in, but as the original user
        self.assertIn(auth.SESSION_KEY, self.client.session)
        self.assertEqual(
            str(self.client.session[auth.SESSION_KEY]), str(self.unauthorized_user.id)
        )
        # Exit user should never get set
        self.assertNotIn("exit_users_pk", self.client.session)

    def test_custom_su_login_url(self):
        """Ensure user is sent to login url following successful login"""
        self.client.login(username="authorized", password="pass")
        with self.settings(SU_LOGIN_REDIRECT_URL="/foo/bar"):
            response = self.client.post(
                reverse("login_as_user", args=[self.destination_user.id])
            )
        self.assertEqual(response.status_code, 302)
        self.assertTrue("/foo/bar" in response["Location"])

    def test_custom_login_action(self):
        """Ensure custom login action is called"""
        self.client.login(username="authorized", password="pass")

        flag = {"called": False}

        def custom_action(request, user):
            flag["called"] = True

        with self.settings(SU_CUSTOM_LOGIN_ACTION=custom_action):
            self.client.post(reverse("login_as_user", args=[self.destination_user.id]))
        self.assertTrue(flag["called"])

    def test_last_login_not_changed(self):
        self.destination_user.last_login = datetime(2000, 1, 1, tzinfo=timezone.utc)
        self.destination_user.save()
        self.client.login(username="authorized", password="pass")
        self.client.post(reverse("login_as_user", args=[self.destination_user.id]))
        self.destination_user = User.objects.get(pk=self.destination_user.pk)
        self.assertEqual(self.destination_user.last_login.date(), date(2000, 1, 1))
        # Check the update_last_login function has been reconnected to the user_logged_in signal
        connections = [
            str(ref[1])
            for ref in auth.user_logged_in.receivers
            if "update_last_login" in str(ref[1])
        ]
        self.assertTrue(connections)

    def test_login_signal_reconnected_following_error(self):
        self.client.login(username="authorized", password="pass")

        def error_action(request, user):
            raise Exception()

        with self.settings(SU_CUSTOM_LOGIN_ACTION=error_action):
            try:
                self.client.post(
                    reverse("login_as_user", args=[self.destination_user.id])
                )
            except Exception:
                pass
        # Check the update_last_login function has been reconnected to the user_logged_in signal
        connections = [
            str(ref[1])
            for ref in auth.user_logged_in.receivers
            if "update_last_login" in str(ref[1])
        ]
        self.assertTrue(connections)


class LoginViewTestCase(SuViewsBaseTestCase):
    def test_get_authorised(self):
        """Load the login page as an authorised user"""
        self.client.login(username="authorized", password="pass")
        response = self.client.get(reverse("su_login"))
        self.assertEqual(response.status_code, 200)

    def test_get_unauthorised(self):
        """Load the login page as an authorised user"""
        self.client.login(username="unauthorized", password="pass")
        response = self.client.get(reverse("su_login"))
        self.assertEqual(response.status_code, 302)

    def test_post_unauthorised(self):
        """Post to the login page as an authorised user"""
        self.client.login(username="unauthorized", password="pass")
        response = self.client.post(reverse("su_login"))
        self.assertEqual(response.status_code, 302)

    def test_post_valid(self):
        """Ensure posting valid data logs the user in"""
        self.client.login(username="authorized", password="pass")
        response = self.client.post(
            reverse("su_login"), data=dict(user=self.destination_user.id)
        )
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            str(self.client.session[auth.SESSION_KEY]), str(self.destination_user.id)
        )

    def test_post_non_existent(self):
        """Ensure posting a non-existent user does not log the user in"""
        self.client.login(username="authorized", password="pass")
        response = self.client.post(reverse("su_login"), data=dict(user="999"))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            str(self.client.session[auth.SESSION_KEY]), str(self.authorized_user.id)
        )

    def test_post_invalid(self):
        """Ensure posting invalid data redisplays the form and does not log the user in"""
        self.client.login(username="authorized", password="pass")
        response = self.client.post(reverse("su_login"), data=dict(user="abc"))
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            str(self.client.session[auth.SESSION_KEY]), str(self.authorized_user.id)
        )


class LogoutViewTestCase(SuViewsBaseTestCase):
    def test_valid_get(self):
        """Ensure user can logout via get"""
        s = self.client.session
        s["exit_users_pk"] = [
            ["1", "django.contrib.auth.backends.ModelBackend"],
        ]
        s.save()
        response = self.client.get(reverse("su_logout"))
        self.assertEqual(response.status_code, 302)
        self.assertIn(auth.SESSION_KEY, self.client.session)
        self.assertEqual(
            str(self.client.session[auth.SESSION_KEY]), str(self.authorized_user.id)
        )

    def test_valid_post(self):
        """Ensure user can logout via post"""
        s = self.client.session
        s["exit_users_pk"] = [
            ["1", "django.contrib.auth.backends.ModelBackend"],
        ]
        s.save()
        response = self.client.post(reverse("su_logout"))
        self.assertEqual(response.status_code, 302)
        self.assertIn(auth.SESSION_KEY, self.client.session)
        self.assertEqual(
            str(self.client.session[auth.SESSION_KEY]), str(self.authorized_user.id)
        )

    def test_no_exit_pk(self):
        """Ensure logout fails if no exit pk present in session"""
        response = self.client.get(reverse("su_logout"))
        self.assertEqual(response.status_code, 400)

    def test_non_existant_exit_pk(self):
        """Ensure logout fails if no exit pk present in session"""
        s = self.client.session
        s["exit_users_pk"] = [
            ["999", "django.contrib.auth.backends.ModelBackend"],
        ]
        s.save()
        response = self.client.get(reverse("su_logout"))
        self.assertEqual(response.status_code, 404)

    def test_redirect_url(self):
        """Ensure logout redirect url setting respected"""
        s = self.client.session
        s["exit_users_pk"] = [
            ["1", "django.contrib.auth.backends.ModelBackend"],
        ]
        s.save()
        with self.settings(SU_LOGOUT_REDIRECT_URL="/foo/bar"):
            response = self.client.get(reverse("su_logout"))
        self.assertEqual(response.status_code, 302)
        self.assertTrue("/foo/bar" in response["Location"])



# -*- coding: utf-8 -*-

from django import template

from ..utils import su_login_callback


register = template.Library()


@register.inclusion_tag("su/login_link.html", takes_context=False)
def login_su_link(user):
    return {"can_su_login": su_login_callback(user)}



from django.conf import settings
from django.contrib import admin
from django.urls import include, path
from django.views.generic import TemplateView


admin.autodiscover()

urlpatterns = [
    path("admin/", admin.site.urls),
    path("su/", include("django_su.urls")),
    path("", TemplateView.as_view(template_name="index.html")),
]

if "ajax_select" in settings.INSTALLED_APPS:
    from ajax_select import urls as ajax_select_urls

    urlpatterns += [
        path("admin/lookups/", include(ajax_select_urls)),
    ]



# -*- coding: utf-8 -*-

from django.contrib.auth import get_user_model


User = get_user_model()


class SuBackend(object):
    supports_inactive_user = False

    def authenticate(self, request=None, su=False, user_id=None, **kwargs):
        if not su:
            return None

        try:
            user = User._default_manager.get(pk=user_id)  # pylint: disable=W0212
        except (User.DoesNotExist, ValueError):
            return None

        return user

    def get_user(self, user_id):
        try:
            return User._default_manager.get(pk=user_id)  # pylint: disable=W0212
        except User.DoesNotExist:
            return None



#!/usr/bin/env python

import codecs
import os
import sys

from setuptools import find_packages, setup


# When creating the sdist, make sure the django.mo file also exists:
if "sdist" in sys.argv or "develop" in sys.argv:
    os.chdir("django_su")
    try:
        from django.core import management

        management.call_command("compilemessages", stdout=sys.stderr, verbosity=1)
    except ImportError:
        if "sdist" in sys.argv:
            raise
    finally:
        os.chdir("..")


def read(*parts):
    file_path = os.path.join(os.path.dirname(__file__), *parts)
    return codecs.open(file_path, encoding="utf-8").read()


setup(
    name="django-su",
    version=read("VERSION"),
    license="MIT License",
    install_requires=[
        "django>=2.2",
    ],
    requires=[
        "Django (>=2.2)",
    ],
    description="Login as any user from the Django admin interface, then switch back when done",
    long_description=read("README.rst"),
    author="Adam Charnock",
    author_email="adam@adamcharnock.com",
    maintainer="Basil Shubin",
    maintainer_email="basil.shubin@gmail.com",
    url="http://github.com/adamcharnock/django-su",
    download_url="https://github.com/adamcharnock/django-su/zipball/master",
    packages=find_packages(exclude=("example*", "*.tests*")),
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
)



import os


DATABASES = {"default": {"ENGINE": "django.db.backends.sqlite3", "NAME": ":memory:"}}

TEMPLATE_DIRS = [
    os.path.join(os.path.dirname(os.path.realpath(__file__)), "test_templates"),
]

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": TEMPLATE_DIRS,
        "APP_DIRS": True,
    },
]

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.sites",
    "django.contrib.sessions",
    "django.contrib.staticfiles",
    "django.contrib.contenttypes",
    "django_su",
    "django.contrib.admin",
]

MIDDLEWARE_CLASSES = [
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
]

MIDDLEWARE = MIDDLEWARE_CLASSES

ROOT_URLCONF = "test_urls"

SITE_ID = 1

STATIC_URL = "/static/"

AUTHENTICATION_BACKENDS = [
    "django.contrib.auth.backends.ModelBackend",
    "django_su.backends.SuBackend",
]



# -*- coding: utf-8 -*-

import warnings

from django.conf import settings
from django.contrib.auth import (
    BACKEND_SESSION_KEY,
    SESSION_KEY,
    authenticate,
    get_user_model,
    login,
)
from django.contrib.auth.decorators import user_passes_test
from django.http import Http404, HttpResponseBadRequest, HttpResponseRedirect
from django.shortcuts import get_object_or_404, render
from django.views.decorators.csrf import csrf_protect
from django.views.decorators.http import require_http_methods

from .forms import UserSuForm
from .utils import custom_login_action, su_login_callback


User = get_user_model()


@csrf_protect
@require_http_methods(["POST"])
@user_passes_test(su_login_callback)
def login_as_user(request, user_id):
    userobj = authenticate(request=request, su=True, user_id=user_id)
    if not userobj:
        raise Http404("User not found")

    exit_users_pk = request.session.get("exit_users_pk", default=[])
    exit_users_pk.append(
        (request.session[SESSION_KEY], request.session[BACKEND_SESSION_KEY])
    )

    maintain_last_login = hasattr(userobj, "last_login")
    if maintain_last_login:
        last_login = userobj.last_login

    try:
        if not custom_login_action(request, userobj):
            login(request, userobj)
        request.session["exit_users_pk"] = exit_users_pk
    finally:
        if maintain_last_login:
            userobj.last_login = last_login
            userobj.save(update_fields=["last_login"])

    if hasattr(settings, "SU_REDIRECT_LOGIN"):
        warnings.warn(
            "SU_REDIRECT_LOGIN is deprecated, use SU_LOGIN_REDIRECT_URL",
            DeprecationWarning,
        )

    return HttpResponseRedirect(getattr(settings, "SU_LOGIN_REDIRECT_URL", "/"))


@csrf_protect
@require_http_methods(["POST", "GET"])
@user_passes_test(su_login_callback)
def su_login(request, form_class=UserSuForm, template_name="su/login.html"):
    form = form_class(request.POST or None)
    if form.is_valid():
        return login_as_user(request, form.get_user().pk)

    return render(
        request,
        template_name,
        {
            "form": form,
        },
    )


def su_logout(request):
    exit_users_pk = request.session.get("exit_users_pk", default=[])
    if not exit_users_pk:
        return HttpResponseBadRequest(("This session was not su'ed into. Cannot exit."))

    user_id, backend = exit_users_pk.pop()

    userobj = get_object_or_404(User, pk=user_id)
    userobj.backend = backend

    if not custom_login_action(request, userobj):
        login(request, userobj)
    request.session["exit_users_pk"] = exit_users_pk

    if hasattr(settings, "SU_REDIRECT_EXIT"):
        warnings.warn(
            "SU_REDIRECT_EXIT is deprecated, use SU_LOGOUT_REDIRECT_URL",
            DeprecationWarning,
        )

    return HttpResponseRedirect(getattr(settings, "SU_LOGOUT_REDIRECT_URL", "/"))
